<?php

namespace Drupal\jsonapi_auto_include\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface;
use Drupal\jsonapi\ResourceType\ResourceTypeRelationship;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Event subscriber to automatically include all relationships in JSON:API responses.
 */
class IncludeAllRelationshipsSubscriber implements EventSubscriberInterface {

  protected $requestStack;
  protected $configFactory;
  protected $entityTypeManager;
  protected $resourceTypeRepository;

  public function __construct(
    RequestStack $request_stack,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager,
    ResourceTypeRepositoryInterface $resource_type_repository
  ) {
    $this->requestStack = $request_stack;
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
    $this->resourceTypeRepository = $resource_type_repository;
  }

  public static function getSubscribedEvents() {
    // Set a higher priority to ensure this subscriber runs before others.
    $events[KernelEvents::REQUEST][] = ['onKernelRequest', 300];
    return $events;
  }

  public function onKernelRequest(RequestEvent $event) {
    $request = $event->getRequest();

    // Get the base path from JSON:API Extras configuration.
    $config = $this->configFactory->get('jsonapi_extras.settings');
    $base_path = '/' . trim($config->get('path_prefix') ?: 'jsonapi', '/');

    // Only proceed if this is a JSON:API request.
    if (strpos($request->getPathInfo(), $base_path . '/') !== 0) {
      return;
    }

    // If 'include' parameter is not set, include all relationships.
    if ($request->query->get('jsonapi_auto_include')) {
      // Extract the resource path without the entity ID or UUID.
      $resource_path = substr($request->getPathInfo(), strlen($base_path));
      $resource_path = trim($resource_path, '/');
      $path_parts = explode('/', $resource_path);

      // Remove the entity ID or UUID if present.
      if (isset($path_parts[2]) && $this->isEntityId($path_parts[2])) {
        array_pop($path_parts);
      }
      $resource_path = '/' . implode('/', $path_parts);

      // Get the matching resource type.
      $resource_type = $this->getResourceTypeByPath($resource_path);
      if (!$resource_type) {
        return;
      }

      // Collect includes recursively.
      $includes = $this->getIncludes($resource_type);

      if (!empty($includes)) {
        // Set the 'include' parameter to include all relationships.
        $request->query->set('include', implode(',', $includes));
      }
    }
  }

  /**
   * Determines if a string is likely an entity ID or UUID.
   *
   * @param string $string
   *   The string to check.
   *
   * @return bool
   *   TRUE if it's an ID or UUID, FALSE otherwise.
   */
  protected function isEntityId($string) {
    // Check if the string is a UUID.
    if (preg_match('/^[0-9a-fA-F-]{36}$/', $string)) {
      return TRUE;
    }
    // Check if the string is a numeric ID.
    if (is_numeric($string)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Retrieves the resource type based on the path.
   *
   * @param string $path
   *   The resource path without the entity ID or UUID.
   *
   * @return \Drupal\jsonapi\ResourceType\ResourceType|null
   *   The resource type or null if not found.
   */
  protected function getResourceTypeByPath($path) {
    $resource_types = $this->resourceTypeRepository->all();
    foreach ($resource_types as $resource_type) {
      if ($resource_type->getPath() === $path) {
        return $resource_type;
      }
    }
    return NULL;
  }

  /**
   * Recursively collects all relationship fields from the resource type.
   *
   * @param \Drupal\jsonapi\ResourceType\ResourceType $resource_type
   *   The JSON:API resource type.
   * @param string $prefix
   *   The prefix for nested relationships.
   * @param array $visited
   *   An array to prevent infinite loops.
   * @param int $depth
   *   The recursion depth.
   *
   * @return array
   *   An array of include paths.
   */
  protected function getIncludes($resource_type, $prefix = '', &$visited = [], $depth = 3) {
    if ($depth <= 0) {
      return [];
    }

    $key = $resource_type->getTypeName();
    if (in_array($key, $visited)) {
      return [];
    }
    $visited[] = $key;

    $includes = [];

    // Construct the configuration name using entity type and bundle.
    $entity_type_id = $resource_type->getEntityTypeId();
    $bundle = $resource_type->getBundle();
    $config_name = 'jsonapi_extras.jsonapi_resource_config.' . $entity_type_id . '--' . $bundle;

    $resource_config = $this->configFactory->get($config_name);
    $disabled_fields = [];
    if ($resource_config) {
      $resource_fields = $resource_config->get('resourceFields') ?: [];
      foreach ($resource_fields as $field_name => $field_config) {
        if (!empty($field_config['disabled'])) {
          $disabled_fields[] = $field_name;
        }
      }
    }

    $fields = $resource_type->getFields();
    foreach ($fields as $field_name => $field) {
      // Only consider relationship fields.
      if ($field instanceof ResourceTypeRelationship) {
        // Get the public name of the field.
        $public_field_name = $field->getPublicName();

        // Exclude fields that are disabled via JSON:API Extras.
        if (in_array($public_field_name, $disabled_fields)) {
          continue;
        }

        // Exclude invalid or reserved fields.
        if (in_array($public_field_name, ['type', 'self'])) {
          continue;
        }

        $include_path = $prefix ? $prefix . '.' . $public_field_name : $public_field_name;
        $includes[] = $include_path;

        // Get related resource types.
        $related_resource_types = $field->getRelatableResourceTypes();
        foreach ($related_resource_types as $related_type) {
          // Recursively get nested includes.
          $nested_includes = $this->getIncludes($related_type, $include_path, $visited, $depth - 1);
          $includes = array_merge($includes, $nested_includes);
        }
      }
    }

    return $includes;
  }

}
